/* Copyright 2012 Tobias Marschall
 *
 * This file is part of CLEVER.
 *
 * CLEVER is free software: you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation, either version 3 of the License, or
 * (at your option) any later version.
 *
 * CLEVER is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with CLEVER.  If not, see <http://www.gnu.org/licenses/>.
 */

#ifndef VARIATIONCALLER_H_
#define VARIATIONCALLER_H_

#include <vector>
#include <algorithm>
#include <limits>

#include <boost/math/distributions/normal.hpp>

#include "Variation.h"
#include "PackedAlignmentPair.h"

/** Class providing methods to call a variation given a set of reads supporting
 *  the same phenomenon. */
class VariationCaller {
private:
	double insert_length_popmean;
	double insert_length_popstddev;
	std::vector<double> log_factorial_table;
	// Comparison class to sort alignment pairs by deviation from popmean
	typedef struct alignment_diff_sort_t {
		double popmean;
		bool operator()(const PackedAlignmentPair* a1, const PackedAlignmentPair* a2) {
			return fabs(((double)a1->getInsertLength()) - popmean) < fabs(((double)a2->getInsertLength()) - popmean);
		}
		alignment_diff_sort_t(double popmean) : popmean(popmean) {}
	} alignment_diff_sort_t;
	// Comparison class to sort alignment pairs by weight (from high to low)
	typedef struct alignment_weight_sort_t {
		double popmean;
		bool operator()(const PackedAlignmentPair* a1, const PackedAlignmentPair* a2) {
			return a1->getWeight() > a2->getWeight();
		}
		alignment_weight_sort_t() {}
	} alignment_weight_sort_t;
	double logFactorial(size_t n);
	/** The logarithm of n choose k. */
	double logBinomial(unsigned int n, unsigned int k);
	/** Given log(x) and log(y), return log(x+y). */
	double logAdd(double log_x, double log_y);
	double pvalueUpperBound(std::vector<const PackedAlignmentPair*>& alignments);
	double log1MinusX(double x);
public:
	typedef struct {
		// start, end, and length of the intersection of all insert segments
		unsigned int insert_start;
		unsigned int insert_end;
		int insert_length;
		// difference between weighted mean insert length and population mean
		double diff;
		// sum of weights of alignments
		double total_weight;
	} additional_stats_t;

	VariationCaller(double insert_length_popmean, double insert_length_popstddev);
	virtual ~VariationCaller() {}

	/** Predicts a variation from a set of (compatible) alignment pairs.
	 *  Parameters begin and end must be iterators yielding pointers to
	 *  instances of PackedAlignmentPair. If "stats" is non-null, additional
	 *  statistics are delivered to that struct. */
	template <typename IteratorType>
	Variation call(IteratorType begin, IteratorType end, additional_stats_t* stats) {
		additional_stats_t additional_stats;
		if (stats == 0) stats = &additional_stats;
		stats->insert_start = 0;
		stats->insert_end = 0;
		double weighted_mean = 0.0;
		stats->total_weight = 0.0;
		std::string chromosome = "";
		for (IteratorType it(begin); it != end; ++it) {
			const PackedAlignmentPair& ap = **it;
			if (it == begin) {
				stats->insert_start = ap.getInsertStart();
				stats->insert_end = ap.getInsertEnd();
				chromosome = ap.getChromosome();
			} else {
				stats->insert_start = std::max(stats->insert_start, ap.getInsertStart());
				stats->insert_end = std::min(stats->insert_end, ap.getInsertEnd());
				assert(chromosome.compare(ap.getChromosome()) == 0);
			}
			weighted_mean += ap.getInsertLength() * ap.getWeight();
			stats->total_weight += ap.getWeight();
		}
		weighted_mean /= stats->total_weight;
		stats->insert_length = ((int)stats->insert_end) - stats->insert_start + 1;
		assert(stats->insert_length >= 0);
		stats->diff = weighted_mean - insert_length_popmean;

		std::vector<const PackedAlignmentPair*> alignments(begin, end);
		double pvalue = pvalueUpperBound(alignments);
		// insertion or deletion?
		int indel_start = 0;
		int indel_end = 0;
		int int_diff = (int)roundf(stats->diff);
		if (int_diff>0) {
			// deletion
			if (stats->insert_length > stats->diff) {
				indel_start = stats->insert_start + (stats->insert_length - int_diff)/2;
				indel_end = indel_start + int_diff;
			} else {
				indel_start = stats->insert_start;
				indel_end = stats->insert_end + 1;
			}
			return Variation(chromosome, indel_start, indel_end, pvalue, Variation::DELETION);
		} else {
			indel_start = ((stats->insert_end+1) + stats->insert_start) / 2;
			return Variation(chromosome, indel_start, -int_diff, pvalue, Variation::INSERTION);
		}
	}
};

#endif /* VARIATIONCALLER_H_ */
